@@ -1,5 +1,41 @@
 # -*- Mode: Python; py-indent-offset: 4 -*-
 import sys
+from copy import *
+
+def get_valid_scheme_definitions(defs):
+    return [x for x in defs if isinstance(x, tuple) and len(x) >= 2]
+
+# New Parameter class, wich emulates a tuple for compatibility reasons
+class Parameter(object):
+    def __init__(self, ptype, pname, pdflt, pnull, prop=None):
+        self.ptype = ptype
+        self.pname = pname
+        self.pdflt = pdflt
+        self.pnull = pnull
+        
+    def __len__(self): return 4
+    def __getitem__(self, i):
+        return (self.ptype, self.pname, self.pdflt, self.pnull)[i]
+
+    def merge(self, old):
+        if old.pdflt is not None:
+            self.pdflt = old.pdflt
+        if old.pnull is not None:
+            self.pnull = old.pnull
+
+# Parameter for property based constructors
+class Property(object):
+    def __init__(self, pname, optional, argname):
+        self.pname = pname
+        self.optional = optional
+        self.argname = argname
+
+    def merge(self, old):
+        if old.optional is not None:
+            self.optional = old.optional
+        if old.argname is not None:
+            self.argname = old.argname
+
 
 class Definition:
     def __init__(self, *args):
@@ -13,6 +49,16 @@ class Definition:
 	"""write out this definition in defs file format"""
 	raise RuntimeError, "this is an abstract class"
 
+    def guess_return_value_ownership(self):
+        "return 1 if caller owns return value"
+        if getattr(self, 'is_constructor_of', False):
+            self.caller_owns_return = True
+        elif self.ret in ('char*', 'gchar*', 'string'):
+            self.caller_owns_return = True
+        else:
+            self.caller_owns_return = False
+
+
 class ObjectDef(Definition):
     def __init__(self, name, *args):
 	self.name = name
@@ -22,9 +68,7 @@ class ObjectDef(Definition):
         self.typecode = None
 	self.fields = []
         self.implements = []
-	for arg in args:
-	    if type(arg) != type(()) or len(arg) < 2:
-		continue
+	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'in-module':
 		self.module = arg[1]
 	    elif arg[0] == 'parent':
@@ -70,9 +114,7 @@ class InterfaceDef(Definition):
 	self.c_name = None
         self.typecode = None
 	self.fields = []
-	for arg in args:
-	    if type(arg) != type(()) or len(arg) < 2:
-		continue
+	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'in-module':
 		self.module = arg[1]
 	    elif arg[0] == 'c-name':
@@ -97,9 +139,7 @@ class EnumDef(Definition):
 	self.c_name = None
         self.typecode = None
 	self.values = []
-	for arg in args:
-	    if type(arg) != type(()) or len(arg) < 2:
-		continue
+	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'in-module':
 		self.in_module = arg[1]
 	    elif arg[0] == 'c-name':
@@ -138,9 +178,7 @@ class BoxedDef(Definition):
         self.copy = None
         self.release = None
 	self.fields = []
-	for arg in args:
-	    if type(arg) != type(()) or len(arg) < 2:
-		continue
+	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'in-module':
 		self.module = arg[1]
 	    elif arg[0] == 'c-name':
@@ -185,9 +223,7 @@ class PointerDef(Definition):
 	self.c_name = None
         self.typecode = None
 	self.fields = []
-	for arg in args:
-	    if type(arg) != type(()) or len(arg) < 2:
-		continue
+	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'in-module':
 		self.module = arg[1]
 	    elif arg[0] == 'c-name':
@@ -222,15 +258,14 @@ class MethodDef(Definition):
         dump = 0
 	self.name = name
 	self.ret = None
+        self.caller_owns_return = None
 	self.c_name = None
         self.typecode = None
 	self.of_object = None
 	self.params = [] # of form (type, name, default, nullok)
         self.varargs = 0
         self.deprecated = None
-	for arg in args:
-	    if type(arg) != type(()) or len(arg) < 2:
-		continue
+	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'of-object':
                 self.of_object = arg[1]
 	    elif arg[0] == 'c-name':
@@ -239,6 +274,8 @@ class MethodDef(Definition):
 		self.typecode = arg[1]
 	    elif arg[0] == 'return-type':
 		self.ret = arg[1]
+            elif arg[0] == 'caller-owns-return':
+                self.caller_owns_return = arg[1] in ('t', '#t')
 	    elif arg[0] == 'parameters':
                 for parg in arg[1:]:
                     ptype = parg[0]
@@ -246,11 +283,12 @@ class MethodDef(Definition):
                     pdflt = None
                     pnull = 0
                     for farg in parg[2:]:
+                        assert isinstance(farg, tuple)
                         if farg[0] == 'default':
                             pdflt = farg[1]
                         elif farg[0] == 'null-ok':
                             pnull = 1
-                    self.params.append((ptype, pname, pdflt, pnull))
+                    self.params.append(Parameter(ptype, pname, pdflt, pnull))
             elif arg[0] == 'varargs':
                 self.varargs = arg[1] in ('t', '#t')
             elif arg[0] == 'deprecated':
@@ -261,13 +299,21 @@ class MethodDef(Definition):
                 dump = 1
         if dump:
             self.write_defs(sys.stderr)
+
+        if self.caller_owns_return is None and self.ret is not None:
+            self.guess_return_value_ownership()
         for item in ('c_name', 'of_object'):
             if self.__dict__[item] == None:
                 self.write_defs(sys.stderr)
                 raise RuntimeError, "definition missing required %s" % (item,)
             
-    def merge(self, old):
+    def merge(self, old, parmerge):
+        self.caller_owns_return = old.caller_owns_return
+        self.varargs = old.varargs
 	# here we merge extra parameter flags accross to the new object.
+        if not parmerge:
+            self.params = deepcopy(old.params)
+            return
 	for i in range(len(self.params)):
 	    ptype, pname, pdflt, pnull = self.params[i]
 	    for p2 in old.params:
@@ -282,6 +328,8 @@ class MethodDef(Definition):
 	    fp.write('  (c-name "' + self.c_name + '")\n')
 	if self.typecode:
 	    fp.write('  (gtype-id "' + self.typecode + '")\n')
+        if self.caller_owns_return:
+	    fp.write('  (caller-owns-return #t)\n')
 	if self.ret:
 	    fp.write('  (return-type "' + self.ret + '")\n')
 	if self.deprecated:
@@ -294,6 +342,8 @@ class MethodDef(Definition):
                 if pnull: fp.write(' (null-ok)')
                 fp.write(')\n')
             fp.write('  )\n')
+	if self.varargs:
+	    fp.write('  (varargs #t)\n')
 	fp.write(')\n\n')
 
 class FunctionDef(Definition):
@@ -303,14 +353,13 @@ class FunctionDef(Definition):
 	self.in_module = None
 	self.is_constructor_of = None
 	self.ret = None
+        self.caller_owns_return = None
 	self.c_name = None
         self.typecode = None
 	self.params = [] # of form (type, name, default, nullok)
         self.varargs = 0
         self.deprecated = None
-	for arg in args:
-	    if type(arg) != type(()) or len(arg) < 2:
-		continue
+	for arg in get_valid_scheme_definitions(args):
 	    if arg[0] == 'in-module':
 		self.in_module = arg[1]
 	    elif arg[0] == 'is-constructor-of':
@@ -321,6 +370,8 @@ class FunctionDef(Definition):
 		self.typecode = arg[1]
 	    elif arg[0] == 'return-type':
 		self.ret = arg[1]
+            elif arg[0] == 'caller-owns-return':
+                self.caller_owns_return = arg[1] in ('t', '#t')
 	    elif arg[0] == 'parameters':
                 for parg in arg[1:]:
                     ptype = parg[0]
@@ -332,7 +383,21 @@ class FunctionDef(Definition):
                             pdflt = farg[1]
                         elif farg[0] == 'null-ok':
                             pnull = 1
-                    self.params.append((ptype, pname, pdflt, pnull))
+                    self.params.append(Parameter(ptype, pname, pdflt, pnull))
+	    elif arg[0] == 'properties':
+                if self.is_constructor_of is None:
+                    print >> sys.stderr, "Warning: (properties ...) "\
+                          "is only valid for constructors"
+                for prop in arg[1:]:
+                    pname = prop[0]
+                    optional = False
+                    argname = pname
+                    for farg in prop[1:]:
+                        if farg[0] == 'optional':
+                            optional = True
+                        elif farg[0] == 'argname':
+                            argname = farg[1]
+                    self.params.append(Property(pname, optional, argname))
             elif arg[0] == 'varargs':
                 self.varargs = arg[1] in ('t', '#t')
             elif arg[0] == 'deprecated':
@@ -343,6 +408,9 @@ class FunctionDef(Definition):
                 dump = 1
         if dump:
             self.write_defs(sys.stderr)
+
+        if self.caller_owns_return is None and self.ret is not None:
+            self.guess_return_value_ownership()
         for item in ('c_name',):
             if self.__dict__[item] == None:
                 self.write_defs(sys.stderr)
@@ -350,14 +418,33 @@ class FunctionDef(Definition):
 
     _method_write_defs = MethodDef.__dict__['write_defs']
 
-    def merge(self, old):
+    def merge(self, old, parmerge):
+        self.caller_owns_return = old.caller_owns_return
+        self.varargs = old.varargs
+        if not parmerge:
+            self.params = deepcopy(old.params)
+            return
 	# here we merge extra parameter flags accross to the new object.
-	for i in range(len(self.params)):
-	    ptype, pname, pdflt, pnull = self.params[i]
-	    for p2 in old.params:
-		if p2[1] == pname:
-		    self.params[i] = (ptype, pname, p2[2], p2[3])
-		    break
+        def merge_param(param):
+	    for old_param in old.params:
+		if old_param.pname == param.pname:
+                    if isinstance(old_param, Property):
+                        # h2def never scans Property's, therefore if
+                        # we have one it was manually written, so we
+                        # keep it.
+                        return deepcopy(old_param)
+                    else:
+                        param.merge(old_param)
+                        return param
+            raise RuntimeError, "could not find %s in old_parameters %r" % (
+                param.pname, [p.pname for p in old.params])
+        try:
+            self.params = map(merge_param, self.params)
+        except RuntimeError:
+            # parameter names changed and we can't find a match; it's
+            # safer to keep the old parameter list untouched.
+            self.params = deepcopy(old.params)
+        
 	if not self.is_constructor_of:
             try:
                 self.is_constructor_of = old.is_constructor_of
@@ -379,17 +466,31 @@ class FunctionDef(Definition):
 	    fp.write('  (c-name "' + self.c_name + '")\n')
 	if self.typecode:
 	    fp.write('  (gtype-id "' + self.typecode + '")\n')
+        if self.caller_owns_return:
+	    fp.write('  (caller-owns-return #t)\n')
 	if self.ret:
 	    fp.write('  (return-type "' + self.ret + '")\n')
 	if self.deprecated:
 	    fp.write('  (deprecated "' + self.deprecated + '")\n')
         if self.params:
-            fp.write('  (parameters\n')
-            for ptype, pname, pdflt, pnull in self.params:
-                fp.write('    \'("' + ptype + '" "' + pname +'"')
-                if pdflt: fp.write(' (default "' + pdflt + '")')
-                if pnull: fp.write(' (null-ok)')
-                fp.write(')\n')
-            fp.write('  )\n')
-	fp.write(')\n\n')
+            if isinstance(self.params[0], Parameter):
+                fp.write('  (parameters\n')
+                for ptype, pname, pdflt, pnull in self.params:
+                    fp.write('    \'("' + ptype + '" "' + pname +'"')
+                    if pdflt: fp.write(' (default "' + pdflt + '")')
+                    if pnull: fp.write(' (null-ok)')
+                    fp.write(')\n')
+                fp.write('  )\n')
+            elif isinstance(self.params[0], Property):
+                fp.write('  (properties\n')
+                for prop in self.params:
+                    fp.write('    \'("' + prop.pname +'"')
+                    if prop.optional: fp.write(' (optional)')
+                    fp.write(')\n')
+                fp.write('  )\n')
+            else:
+                assert False, "strange parameter list %r" % self.params[0]
+	if self.varargs:
+	    fp.write('  (varargs #t)\n')
 
+	fp.write(')\n\n')
